#include "common.hpp"

#include "catch_config.hpp"

#include "catch.hpp"

#include "../HeapUtils.hpp"

SCENARIO ("[2] parent() should always return the parent's index", "[parent]")
{
  size_t expected_parent = 0;
  for (size_t child = 1; child < 20; ++child)
  {
    REQUIRE (HeapUtils::detail::parent (child) == expected_parent);
    if (!(child & 1))
    {
      ++expected_parent;
    }
  }
}

SCENARIO ("[2] left_child() should always return the left child's index",
          "[left]")
{
  for (size_t left_child = 1; left_child < 30; left_child += 2)
  {
    size_t parent = HeapUtils::detail::parent (left_child);
    REQUIRE (HeapUtils::detail::left_child (parent) == left_child);
  }
}

SCENARIO ("[2] right_child() should always return the right child's index",
          "[right]")
{
  for (size_t right_child = 2; right_child < 30; right_child += 2)
  {
    size_t parent = HeapUtils::detail::parent (right_child);
    REQUIRE (HeapUtils::detail::right_child (parent) == right_child);
  }
}

SCENARIO ("[2] last_interior() should always return the correct index",
          "[last_interior]")
{
  size_t const MAX_SIZE = 100;
  for (size_t size = 2; size <= MAX_SIZE; ++size)
  {
    size_t last = size - 1;
    size_t last_interior = HeapUtils::detail::parent (last);
    CAPTURE (size);
    REQUIRE (HeapUtils::detail::last_interior (size) == last_interior);
  }
}

SCENARIO ("max_child() should always return the \"greater\" child index",
          "[max_child]")
{
  size_t pos = 0;
  size_t left_child = 1;
  size_t right_child = 2;
  std::vector<int> v;
  std::less<> less_than;
  std::greater<> greater_than;
  v.reserve (3);
  GIVEN ("An index with only a left child")
  {
    THEN ("[1] we should return the left child")
    {
      v = {3, 2};
      CAPTURE (v, pos);
      REQUIRE (HeapUtils::detail::max_child (v, pos, v.size(), less_than) == left_child);
    }
  }

  GIVEN ("An index with only both children and the left is greater")
  {
    WHEN ("We use a less<> comparator")
    {
      THEN ("[0.5] we should return the left child")
      {
        v = {3, 2, 1};
        CAPTURE (v, pos);
        REQUIRE (HeapUtils::detail::max_child (v, pos, v.size(), less_than) ==
                 left_child);
      }
    }
    WHEN ("We use a greater<> comparator")
    {
      THEN ("[0.5] we should return the left child")
      {
        v = {1, 2, 3};
        CAPTURE (v, pos);
        REQUIRE (HeapUtils::detail::max_child (v, pos, v.size(), greater_than) ==
                 left_child);
      }
    }
  }
  GIVEN ("An index with only both children and the right is greater")
  {
    WHEN ("We use a less<> comparator")
    {
      THEN ("[0.5] we should return the right child")
      {
        v = {3, 1, 2};
        CAPTURE (v, pos);
        REQUIRE (HeapUtils::detail::max_child (v, pos, v.size(), less_than) ==
                 right_child);
      }
    }
    WHEN ("We use a greater<> comparator")
    {
      THEN ("[0.5] we should return the right child")
      {
        v = {1, 3, 2};
        CAPTURE (v, pos);
        REQUIRE (HeapUtils::detail::max_child (v, pos, v.size(), greater_than) ==
                 right_child);
      }
    }
  }
  GIVEN ("An index with only both children and they are equal")
  {
    WHEN ("We use a less<> comparator")
    {
      THEN ("[0.5] we should return the left child")
      {
        v = {2, 1, 1};
        CAPTURE (v, pos);
        REQUIRE (HeapUtils::detail::max_child (v, pos, v.size(), less_than) ==
                 left_child);
      }
    }
    WHEN ("We use a greater<> comparator")
    {
      THEN ("[0.5] we should return the left child")
      {
        v = {1, 2, 2};
        CAPTURE (v, pos);
        REQUIRE (HeapUtils::detail::max_child (v, pos, v.size(), greater_than) ==
                 left_child);
      }
    }
  }
}

namespace solution
{

template<typename T, typename Comp>
void
push_heap (std::vector<T>& v, Comp& comp)
{
  std::push_heap (v.begin (), v.end (), comp);
}

template<typename T, typename Comp>
bool
is_heap (std::vector<T>& v, Comp& comp)
{
  return std::is_heap (v.begin (), v.end (), comp);
}

} // end namespace solution

std::vector<Counter<int>>
random_vector (size_t size, unsigned int seed)
{
  std::vector<Counter<int>> v;
  for (size_t i = 0; i < size; ++i)
  {
    v.emplace_back (int(i + 1));
  }
  std::minstd_rand rng (seed);
  std::shuffle (v.begin (), v.end (), rng);
  return v;
}

unsigned int
seed_for (size_t count)
{
  return (unsigned int)((8ULL << count) * (count << 5 ^ 0xDEADULL));
}

size_t
ancestor (size_t loc, size_t anc)
{
  if (loc < anc)
    return false;
  if (loc == anc)
    return true;
  return ancestor (HeapUtils::detail::parent (loc), anc);
}

template<typename T>
void
print_heap_impl (std::vector<T> const& h, size_t index, size_t depth,
                 std::ostream& oss, size_t idx)
{
  if (index >= h.size ())
  {
    return;
  }
  print_heap_impl (h, HeapUtils::detail::right_child (index), depth + 4, oss,
                   idx);
  for (size_t i = 0; i < depth; ++i)
    oss << ' ';
  if (idx == index)
    oss << '[';
  oss << h[index];
  if (idx == index)
    oss << ']';
  oss << '\n';
  print_heap_impl (h, HeapUtils::detail::left_child (index), depth + 4, oss,
                   idx);
  return;
}

template<typename T>
std::string
print_heap (std::vector<T> const& h, size_t idx = -1)
{
  std::ostringstream oss;
  oss << '\n';
  print_heap_impl (h, 0, 0, oss, idx);
  return oss.str ();
}

template<typename Container>
auto
sorted (Container c)
{
  std::sort (c.begin (), c.end ());
  return c;
}

template<typename T, typename Comp>
bool
is_heap_from (std::vector<T> const& v, size_t index, Comp& comp)
{
  if (index > HeapUtils::detail::last_interior (v.size ()))
    return true;
  size_t left = HeapUtils::detail::left_child (index);
  size_t right = HeapUtils::detail::right_child (index);
  bool is_heap = true;
  if (left < v.size ())
  {
    is_heap = !comp (v[index], v[left]) && is_heap_from (v, left, comp);
    if (is_heap)
    {
      if (right < v.size ())
      {
        is_heap = !comp (v[index], v[right]) && is_heap_from (v, right, comp);
      }
    }
  }
  return is_heap;
}

SCENARIO ("sift_down properly down-heaps a value at an index into place",
          "[sift_down]")
{
  WHEN ("[5] we use the less than comparator")
  {
    std::less<> less_than;
    std::vector<int> heap{4, 12, 3, 11, 5, 13, 2, 10,
                          6, 14, 1, 9,  7, 15, 8, 16};
    for (size_t i = heap.size (); i > 0;)
    {
      int val = heap[--i];
      INFO ("Before operation");
      CAPTURE (i, print_heap (heap, i));
      HeapUtils::detail::sift_down (heap, i, heap.size (), less_than);
      INFO ("After operation");
      CAPTURE (print_heap (heap, i));
      REQUIRE (is_heap_from (heap, i, less_than));
      size_t newLoc = std::distance (
        heap.begin (), std::find (heap.begin () + i, heap.end (), val));
      bool sameIndex = newLoc == i;
      bool isAncestor = ancestor (newLoc, i);
      CAPTURE (sameIndex, isAncestor);
      REQUIRE ((sameIndex || isAncestor));
    }
  }
  WHEN ("[5] we use the greater than comparator")
  {
    std::greater<> greater_than;
    std::vector<int> heap{4, 12, 3, 11, 5, 13, 2, 10,
                          6, 14, 1, 9,  7, 15, 8, 16};
    for (size_t i = heap.size (); i > 0;)
    {
      int val = heap[--i];
      INFO ("Before operation");
      CAPTURE (i, print_heap (heap, i));
      HeapUtils::detail::sift_down (heap, i, heap.size (), greater_than);
      INFO ("After operation");
      CAPTURE (print_heap (heap, i));
      REQUIRE (is_heap_from (heap, i, greater_than));
      size_t newLoc = std::distance (
        heap.begin (), std::find (heap.begin () + i, heap.end (), val));
      bool sameIndex = newLoc == i;
      bool isAncestor = ancestor (newLoc, i);
      CAPTURE (sameIndex, isAncestor);
      REQUIRE ((sameIndex || isAncestor));
    }
  }
}

SCENARIO ("sift_up properly up-heaps a value at an index into place",
          "[sift_up]")
{
  WHEN ("[5] we use the less than comparator")
  {
    std::less<> less_than;
    std::vector<int> heap{4, 12, 3, 11, 5, 13, 2, 10,
                          6, 14, 1, 9,  7, 15, 8, 16};
    std::vector<int> sorted_heap = sorted (heap);
    for (size_t i = 0; i < heap.size (); ++i)
    {
      int val = heap[i];
      INFO ("Before operation");
      CAPTURE (i, print_heap (heap, i));
      HeapUtils::detail::sift_up (heap, i, less_than);
      INFO ("After operation");
      CAPTURE (print_heap (heap, i));
      {
        INFO ("everything above us is still a heap");
        REQUIRE (std::is_heap (heap.begin (), heap.begin () + i, less_than));
      }
      {
        INFO ("value was propagated up");
        size_t newLoc = std::distance (
          heap.begin (), std::find (heap.begin (), heap.begin () + i, val));
        CAPTURE (newLoc);
        REQUIRE ((newLoc == i || ancestor (i, newLoc)));
      }
      {
        INFO ("No loss of elements");
        REQUIRE (sorted (heap) == sorted_heap);
      }
    }
  }
  WHEN ("[5] we use the greater than comparator")
  {
    std::greater<> greater_than;
    std::vector<int> heap{4, 12, 3, 11, 5, 13, 2, 10,
                          6, 14, 1, 9,  7, 15, 8, 16};
    std::vector<int> sorted_heap = sorted (heap);
    for (size_t i = 0; i < heap.size (); ++i)
    {
      int val = heap[i];
      INFO ("Before operation");
      CAPTURE (i, print_heap (heap, i));
      HeapUtils::detail::sift_up (heap, i, greater_than);
      INFO ("After operation");
      CAPTURE (print_heap (heap, i));
      {
        INFO ("everything above us is still a heap");
        REQUIRE (std::is_heap (heap.begin (), heap.begin () + i, greater_than));
      }
      {
        INFO ("value was propagated up");
        size_t newLoc = std::distance (
          heap.begin (), std::find (heap.begin (), heap.begin () + i, val));
        CAPTURE (newLoc);
        REQUIRE ((newLoc == i || ancestor (i, newLoc)));
      }
      {
        INFO ("No loss of elements");
        REQUIRE (sorted (heap) == sorted_heap);
      }
    }
  }
}

SCENARIO ("make_heap properly \"heapifies\" a vector", "[make_heap]")
{
  WHEN ("[3] we use the less than comparator")
  {
    std::less<> less_than;
    auto heap = random_vector (10, 1337);
    auto sorted_heap = sorted (heap);
    HeapUtils::make_heap (heap, less_than);

    {
      INFO ("there were no loss of elements");
      REQUIRE (sorted_heap == sorted (heap));
    }
    {
      INFO ("the structure must be a heap");
      CAPTURE (print_heap (heap));
      REQUIRE (std::is_heap (heap.begin (), heap.end (), less_than));
    }
  }

  WHEN ("[3] we use the greater than comparator")
  {
    std::greater<> greater_than;
    auto heap = random_vector (10, 1337);
    auto sorted_heap = sorted (heap);
    HeapUtils::make_heap (heap, greater_than);

    {
      INFO ("there were no loss of elements");
      REQUIRE (sorted_heap == sorted (heap));
    }
    {
      INFO ("the structure must be a heap");
      CAPTURE (print_heap (heap));
      REQUIRE (std::is_heap (heap.begin (), heap.end (), greater_than));
    }
  }
}

SCENARIO ("push_heap inserts back() into the heap represented by the vector",
          "[push_heap]")
{
  std::vector<int> data{4, 12, 3, 11, 5, 13, 2, 10, 6, 14, 1, 9, 7, 15, 8, 16};
  std::vector<int> expected, actual;
  WHEN ("[2] we use the less than comparator")
  {
    std::less<> less_than;
    for (int value : data)
    {
      actual.push_back (value);
      expected.push_back (value);
      solution::push_heap (expected, less_than);
      HeapUtils::push_heap (actual, less_than);
      INFO ("After inserting " << value);
      REQUIRE (expected == actual);
    }
  }

  WHEN ("[2] we use the greater than comparator")
  {
    std::greater<> greater_than;
    for (int value : data)
    {
      actual.push_back (value);
      expected.push_back (value);
      solution::push_heap (expected, greater_than);
      HeapUtils::push_heap (actual, greater_than);
      INFO ("After inserting " << value);
      REQUIRE (expected == actual);
    }
  }
}

SCENARIO ("pop_heap removes front() from the heap represented by the vector",
          "[pop_heap]")
{
  auto data = random_vector(17, 8);
  WHEN ("[2] we use the less than comparator")
  {
    std::less<> less_than;
    auto expected = data;
    std::make_heap (expected.begin(), expected.end(), less_than);
    auto actual = expected;

    while (expected.size () > 0)
    {
      INFO ("Before pop_heap");
      CAPTURE (actual);
      std::pop_heap (expected.begin(), expected.end(), less_than);
      HeapUtils::pop_heap (actual, less_than);
      INFO ("After pop_heap");
      CAPTURE (print_heap(expected), print_heap(actual));
      REQUIRE (expected.back() == actual.back());
      expected.pop_back ();
      actual.pop_back ();
      REQUIRE (std::is_heap (actual.begin (), actual.end (), less_than));
    }
  }

  WHEN ("[2] we use the greater than comparator")
  {
    std::greater<> greater_than;
    auto expected = data;
    std::make_heap (expected.begin(), expected.end(), greater_than);
    auto actual = expected;

    while (expected.size () > 0)
    {
      INFO ("Before pop_heap");
      CAPTURE (actual);
      std::pop_heap (expected.begin(), expected.end(), greater_than);
      HeapUtils::pop_heap (actual, greater_than);
      INFO ("After pop_heap");
      CAPTURE (print_heap(expected), print_heap(actual));
      REQUIRE (expected.back() == actual.back());
      expected.pop_back ();
      actual.pop_back ();
      REQUIRE (std::is_heap (actual.begin (), actual.end (), greater_than));
    }
  }
}

SCENARIO ("[4] is_heap correctly identifies whether the vector is a heap",
          "[is_heap]")
{
  std::less<> less_than;
  std::greater<> greater_than;
  auto heap = random_vector (7, 1337);
  std::sort (heap.begin (), heap.end ());
  do
  {
    {
      INFO ("When using less_than");
      Counters::reset ();
      bool expected = std::is_heap (heap.begin (), heap.end (), less_than);
      size_t expected_count = Counters::compares;
      Counters::reset ();
      bool actual = HeapUtils::is_heap (heap, less_than);
      size_t actual_count = Counters::compares;
      if (expected_count != actual_count)
      {
        WARN ("Comparisons [expected]=" << expected_count
                                        << "  vs  [actual]=" << actual_count);
      }
      CAPTURE (heap);
      REQUIRE (actual == expected);
    }
    {
      INFO ("When using greater_than");
      Counters::reset ();
      bool expected = std::is_heap (heap.begin (), heap.end (), greater_than);
      size_t expected_count = Counters::compares;
      Counters::reset ();
      bool actual = HeapUtils::is_heap (heap, greater_than);
      size_t actual_count = Counters::compares;
      if (expected_count != actual_count)
      {
        WARN ("Comparisons [expected]=" << expected_count
                                        << "  vs  [actual]=" << actual_count);
      }
      CAPTURE (heap);
      REQUIRE (actual == expected);
    }
  } while (std::next_permutation (heap.begin (), heap.end ()));
}
